// addrspace.cc 
//	Routines to manage address spaces (executing user programs).
//
//	In order to run a user program, you must:
//
//	1. link with the -n -T 0 option 
//	2. run coff2noff to convert the object file to Nachos format
//		(Nachos object code format is essentially just a simpler
//		version of the UNIX executable object code format)
//	3. load the NOFF file into the Nachos file system
//		(if you are using the "stub" file system, you
//		don't need to do this last step)
//
// Copyright (c) 1992-1996 The Regents of the University of California.
// All rights reserved.  See copyright.h for copyright notice and limitation 
// of liability and disclaimer of warranty provisions.

#include "copyright.h"
#include "main.h"
#include "machine.h"
#include "addrspace.h"
#include "noff.h"
#include "synchbitmap.h"
#include "table.h"
#include "translate.h"
#include "openfile.h"
#include "utility.h"

//----------------------------------------------------------------------
// SwapHeader
// 	Do little endian to big endian conversion on the bytes in the 
//	object file header, in case the file was generated on a little
//	endian machine, and we're now running on a big endian machine.
//----------------------------------------------------------------------

static void 
SwapHeader (NoffHeader *noffH)
{
    noffH->noffMagic = WordToHost(noffH->noffMagic);
    noffH->code.size = WordToHost(noffH->code.size);
    noffH->code.virtualAddr = WordToHost(noffH->code.virtualAddr);
    noffH->code.inFileAddr = WordToHost(noffH->code.inFileAddr);
#ifdef RDATA
    noffH->readonlyData.size = WordToHost(noffH->readonlyData.size);
    noffH->readonlyData.virtualAddr = 
           WordToHost(noffH->readonlyData.virtualAddr);
    noffH->readonlyData.inFileAddr = 
           WordToHost(noffH->readonlyData.inFileAddr);
#endif 
    noffH->initData.size = WordToHost(noffH->initData.size);
    noffH->initData.virtualAddr = WordToHost(noffH->initData.virtualAddr);
    noffH->initData.inFileAddr = WordToHost(noffH->initData.inFileAddr);
    noffH->uninitData.size = WordToHost(noffH->uninitData.size);
    noffH->uninitData.virtualAddr = WordToHost(noffH->uninitData.virtualAddr);
    noffH->uninitData.inFileAddr = WordToHost(noffH->uninitData.inFileAddr);

}

//----------------------------------------------------------------------
// AddrSpace::AddrSpace
//      Create an address space to run a user program.
//      Load the program from a file "executable", and set everything
//      up so that we can start executing user instructions.
//
//      Assumes that the object code file is in NOFF format.
//
//      "executable" is the file containing the object code to load into memory
//      "map" is a bitmap that can be used to determine which frames
//         are available
//      "myid" is the unique identifier for this address space
//
//      If the address space cannot be initialized because of lack
//      of space in primary memory, the constructor sets the id
//      of the address space to -1.
//----------------------------------------------------------------------

AddrSpace::AddrSpace(OpenFile *executable,SynchBitmap* map,int myid)
{
    NoffHeader noffH;
    int i;
    int paddr;
    int pagenum;
    int numbytes;

    // initialize object
    memMap = map;
    id = myid;
    pageTable = (TranslationEntry *)0;
    numPages = 0;
    openFiles = (Table *)0;

    executable->ReadAt((char *)&noffH, sizeof(noffH), 0);
    if ((noffH.noffMagic != NOFFMAGIC) && 
                (WordToHost(noffH.noffMagic) == NOFFMAGIC))
        SwapHeader(&noffH);
    
    if (noffH.noffMagic != NOFFMAGIC) {
      // doesn't look like executable is a noff file
      id = -1;
      return;
    }

    // how big is the address space?
    // the following assumes that each segment is page-aligned
    numPages = divRoundUp(noffH.code.size,PageSize) + 
      divRoundUp(noffH.initData.size,PageSize) + 
      divRoundUp(noffH.uninitData.size,PageSize) +
      divRoundUp(UserStackSize,PageSize); 
#ifdef RDATA
    numPages += divRoundUp(noffH.readonlyData.size,PageSize);
#endif

    DEBUG(dbgAddr, "Initializing address space: " << numPages << " pages."); 
#ifdef RDATA
    DEBUG(dbgAddr, "code = " << noffH.code.size <<  
                   " readonly = " << noffH.readonlyData.size <<
                   " init = " << noffH.initData.size <<
                   " uninit = " << noffH.uninitData.size);
#else
    DEBUG(dbgAddr, "code = " << noffH.code.size <<  
                   " init = " << noffH.initData.size <<
                   " uninit = " << noffH.uninitData.size);
#endif

    // Set up a translation table for this process
    pageTable = new TranslationEntry[numPages];
    ASSERT(pageTable != (TranslationEntry*)0);
    // mark all entries as invalid initially
    for (i = 0; i < numPages; i++) {
      pageTable[i].valid = FALSE;
    }
    // initialize the page table
    for (i = 0; i < numPages; i++) {
      pageTable[i].virtualPage = i;
      pageTable[i].physicalPage = memMap->FindAndSet();
      if (pageTable[i].physicalPage < 0) {
	// we ran out of free frames
	// until we have paging, this means we cannot run the process
	id = -1;  // mark this as a bad address space
	return;
      }
      pageTable[i].valid = TRUE;
      pageTable[i].use = FALSE;
      pageTable[i].dirty = FALSE;
      // this will be changed later for read-only pages
      pageTable[i].readOnly = FALSE; 
    }

    // zero out the entire address space, to zero the unitialized data segment 
    // and the stack segment and any partially-filled pages
    // do this one page at a time
    for(i=0;i<numPages;i++) {
      bzero(&(kernel->machine->mainMemory[pageTable[i].physicalPage*PageSize]),
            PageSize);
    }

    // create an open file table
    openFiles = new Table(MaxOpenFiles);
    ASSERT(openFiles != (Table*)0);

    // initialize the code segment
    while(noffH.code.size > 0) {
      if (PageSize < noffH.code.size) {
	numbytes = PageSize;
      } else {
	numbytes = noffH.code.size;
      }
      pagenum = noffH.code.virtualAddr / PageSize;
      ASSERT(pagenum >= 0);
      ASSERT(pagenum < numPages);
      if (Translate(noffH.code.virtualAddr,&paddr,FALSE) != OK) {
	ASSERT(FALSE);
      }
      ASSERT(paddr >= 0);
      // load the code from the file
      executable->ReadAt(&(kernel->machine->mainMemory[paddr]),
                        numbytes, noffH.code.inFileAddr);
      DEBUG(dbgAddr, "Loaded code segment page. VA: " << noffH.code.virtualAddr << " bytes: " << numbytes); 
      pageTable[pagenum].use = FALSE; // undo side-effect of Translate
      pageTable[pagenum].readOnly = TRUE;
      noffH.code.size -= numbytes;
      noffH.code.inFileAddr += numbytes;
      noffH.code.virtualAddr += numbytes;
    }
    // initialize the read only data segment
#ifdef RDATA
    while(noffH.readonlyData.size > 0) {
      if (PageSize < noffH.readonlyData.size) {
	numbytes = PageSize;
      } else {
	numbytes = noffH.readonlyData.size;
      }
      pagenum = noffH.readonlyData.virtualAddr / PageSize;
      ASSERT(pagenum >= 0);
      ASSERT(pagenum < numPages);
      if (Translate(noffH.readonlyData.virtualAddr,&paddr,FALSE) != OK) {
	ASSERT(FALSE);
      }
      ASSERT(paddr >= 0);
      // load the read-only data from the file
      executable->ReadAt(&(kernel->machine->mainMemory[paddr]),
                        numbytes, noffH.readonlyData.inFileAddr);
      DEBUG(dbgAddr, "Loaded read-only data segement page. VA: " << noffH.readonlyData.virtualAddr << " bytes: " << numbytes); 
      pageTable[pagenum].use = FALSE;  // undo side-effect of Translate
      pageTable[pagenum].readOnly = TRUE;
      noffH.readonlyData.size -= numbytes;
      noffH.readonlyData.inFileAddr += numbytes;
      noffH.readonlyData.virtualAddr += numbytes;
    }
#endif
    // initialize the initialized data segment
    while(noffH.initData.size > 0) {
      if (PageSize < noffH.initData.size) {
	numbytes = PageSize;
      } else {
	numbytes = noffH.initData.size;
      }
      pagenum = noffH.initData.virtualAddr / PageSize;
      ASSERT(pagenum >= 0);
      ASSERT(pagenum < numPages);
      if (Translate(noffH.initData.virtualAddr,&paddr,FALSE) != OK) {
	ASSERT(FALSE);
      }
      ASSERT(paddr >= 0);
      // zero out the page if it will only be partially filled
      if (numbytes < PageSize) {
	bzero(&(kernel->machine->mainMemory[paddr]),PageSize);
      }
      // initialized the data from the file
      executable->ReadAt(&(kernel->machine->mainMemory[paddr]),
                        numbytes, noffH.initData.inFileAddr);
      DEBUG(dbgAddr, "Loaded initialized data segment page. VA: "<< noffH.initData.virtualAddr << " bytes: " << numbytes); 
      pageTable[pagenum].use = FALSE;  // undo side-effect of Translate
      noffH.initData.size -= numbytes;
      noffH.initData.inFileAddr += numbytes;
      noffH.initData.virtualAddr += numbytes;
    }
}


//----------------------------------------------------------------------
// AddrSpace::~AddrSpace
// 	Dealloate an address space.
//----------------------------------------------------------------------

AddrSpace::~AddrSpace()
{
  int i;
  OpenFile* f;

  if (pageTable != (TranslationEntry*)0) {
    // free up the frames we are using
    for(i=0;i<numPages;i++) {
      if (pageTable[i].valid) {
	memMap->Clear(pageTable[i].physicalPage);
      }
    }
    // delete the page table
    delete pageTable;
  }

  if (openFiles != (Table*)0) {
    // close any open files associated with this address space
    for(i=0;i<MaxOpenFiles;i++) {
      f = (OpenFile*)(openFiles->Fetch(i));
      if (f != (OpenFile*)0) delete f;
    }
    // delete the openFiles table itself
    delete openFiles;
  }
}

//----------------------------------------------------------------------
// AddrSpace::InitRegisters
// 	Set the initial values for the user-level register set.
//
// 	We write these directly into the "machine" registers, so
//	that we can immediately jump to user code.  Note that these
//	will be saved/restored into the currentThread->userRegisters
//	when this thread is context switched out.
//----------------------------------------------------------------------

void
AddrSpace::InitRegisters()
{
    Machine *machine = kernel->machine;
    int i;

    for (i = 0; i < NumTotalRegs; i++)
	machine->WriteRegister(i, 0);

    // Initial program counter -- must be location of "Start"
    machine->WriteRegister(PCReg, 0);	

    // Need to also tell MIPS where next instruction is, because
    // of branch delay possibility
    // Since instructions occupy four bytes each, the next instruction
    // after start will be at virtual address four.
    machine->WriteRegister(NextPCReg, 4);

   // Set the stack register to the end of the address space, where we
   // allocated the stack; but subtract off a bit, to make sure we don't
   // accidentally reference off the end!
    machine->WriteRegister(StackReg, numPages * PageSize - 16);
}

//----------------------------------------------------------------------
// AddrSpace::SaveState
// 	On a context switch, save any machine state, specific
//	to this address space, that needs saving.
//
//	For now, don't need to save anything!
//----------------------------------------------------------------------

void AddrSpace::SaveState() 
{}

//----------------------------------------------------------------------
// AddrSpace::RestoreState
// 	On a context switch, restore the machine state so that
//	this address space can run.
//
//      For now, tell the machine where to find the page table.
//----------------------------------------------------------------------

void AddrSpace::RestoreState() 
{
    kernel->machine->pageTable = pageTable;
    kernel->machine->pageTableSize = numPages;
}


//-------------------------------------------------------------------
// AddrSpace::Translate
//      map a virtual address to a physical one
//
//   "vaddr" is the virtual address to translate.
//   "*paddr" is the resulting physical address
//   "write" should be true if the translation is for an update,
//            false otherwise
//
//   return value is a TranslationStatus indicating whether or
//     not the translation was successful.
//   a physical address is written into *paddr only if the
//    return value is "OK"
//
//-------------------------------------------------------------------

TranslationStatus AddrSpace::Translate(int vaddr,int *paddr, bool write)
{
  int pagenum;
  int offset;
  int framenum;

  ASSERT(vaddr >= 0);
  ASSERT(paddr != (int *)0);
  pagenum = vaddr/PageSize;
  if (pagenum >= numPages) {
    return(AddressError);
  }
  if (!pageTable[pagenum].valid) {
    return(PageFault);
  }
  if (write && pageTable[pagenum].readOnly) {
    return(ProtectionError);
  }
  framenum = pageTable[pagenum].physicalPage;
  if ((framenum >= NumPhysPages) || (framenum < 0)) {
    return(BusError);
  }
  pageTable[pagenum].use = TRUE;
  if (write) {
    pageTable[pagenum].dirty = TRUE;
  }
  offset = vaddr%PageSize;
  *paddr = framenum*PageSize + offset;
  ASSERT(*paddr < MemorySize);
  ASSERT(*paddr >= 0);
  return(OK);
}

//----------------------------------------------------------------------
// AddrSpace::ReadMem
//      Read "size" (1, 2, or 4) bytes of virtual memory at virtual
//      address "addr" into *value
//
//  Returns a TranslationStatus to indicate whether the required
//    virtual to physical translation was successful.
//  Places data into *value only if the return value is "OK"
//----------------------------------------------------------------------

TranslationStatus
AddrSpace::ReadMem(int addr, int size, int *value)
{
  int physicalAddress;
  TranslationStatus status;

  ASSERT(value != (int *)0);
  ASSERT(addr >= 0);
  status = Translate(addr, &physicalAddress, FALSE);
  if (status != OK) {
    return(status);
  }
  switch (size) {
  case 1:
    *value = kernel->machine->mainMemory[physicalAddress];
    break;
    
  case 2:
    if (addr & 0x1) {
      // alignment error
      return(AddressError);
    }
    *value = ShortToHost(*(unsigned short *) &kernel->machine->mainMemory[physicalAddress]);
    break;
    
  case 4:
    if (addr & 0x3) {
      // alignment error
      return(AddressError);
    }
    *value = WordToHost(*(unsigned int *) &kernel->machine->mainMemory[physicalAddress]);
    break;
      
  default: ASSERT(FALSE);
  }
  return(OK);
}

//----------------------------------------------------------------------
// AddrSpace::WriteMem
//      Write "value", which is "size" bytes, to virtual address "addr"
//  Returns a TranslationStatus to indicate whether the required
//    virtual to physical translation was successful.
//  Writes data to addr only if the return value is "OK"
//----------------------------------------------------------------------

TranslationStatus
AddrSpace::WriteMem(int addr, int size, int value)
{
  int physicalAddress;
  TranslationStatus status;

  ASSERT(addr >= 0);
  status = Translate(addr, &physicalAddress,TRUE);
  if (status != OK) {
    return(status);
  }
  switch (size) {
  case 1:
    kernel->machine->mainMemory[physicalAddress] =
      (unsigned char)(value & 0xff);
    break;
    
  case 2:
    if (addr & 0x1) {
      // alignment error
      return(AddressError);
    }
    *(unsigned short *) &(kernel->machine->mainMemory[physicalAddress]) = 
      ShortToMachine((unsigned short)(value & 0xffff));
    break;
    
  case 4:
    if (addr & 0x3) {
      // alignment error
      return(AddressError);
    }
    *(unsigned int *) &(kernel->machine->mainMemory[physicalAddress]) = 
      WordToMachine((unsigned int)value);
    break;
      
  default: ASSERT(FALSE);
  }
  return(OK);
}

//------------------------------------------------------------------
// AddrSpace::AddOpenFile
//     add a new open file to the list of open files for this
//     address space, and assign an unused file identifier to it
//
//     "file" is a pointer to an already-created OpenFile object
//
//     returns an integer, which is the file identifier
//     a negative return value indicates that a new file identifier
//     cannot be created because there are too many in use
//
//     Note that no two files open concurrently
//     in the same address space will
//     have the same file identifier.  However, file identifiers
//      are reused after files are closed.
//     Note also that two opens on the same file will result
//      in two distinct file identifiers (like UNIX).
//------------------------------------------------------------------

int
AddrSpace::AddOpenFile(OpenFile* file)
{
  int fileId;
  
  fileId = openFiles->Insert((void*)file);
  if (fileId < 0) {
    // open file table is full
    DEBUG(dbgAddr, "Attempting to add open file: file table full."); 
    return(fileId);
  } else {
    ASSERT(fileId < MaxOpenFiles);
    DEBUG(dbgAddr, "Adding open file with fileId " << fileId+2); 
    return(fileId+2);
  }
}

//------------------------------------------------------------------
// AddrSpace::DeleteOpenFile
//     if a file with the specified identifier is on the list then
//      return it and delete it from the list, else return a null
//      pointer
//
//     "fId" is the identifier of the desired file
//------------------------------------------------------------------

OpenFile* 
AddrSpace::DeleteOpenFile(int fId)
{
  OpenFile* temp;

  if ((fId < 2) || (fId >= MaxOpenFiles+2)) {
    return((OpenFile *)0);
  }
  DEBUG(dbgAddr, "Deleting open file with fileId " << fId); 
  temp = (OpenFile*)(openFiles->Fetch(fId-2));
  openFiles->Remove(fId-2);
  return(temp);
}

//------------------------------------------------------------------
// AddrSpace::GetOpenFile
//     if a file with the specified identifier is on the list then
//      return it, else return a null pointer
//     This is like AddrSpace::DeleteOpenFile, except the open file
//     is not deleted.
//
//     "fId" is the identifier of the desired file
//------------------------------------------------------------------

OpenFile* 
AddrSpace::GetOpenFile(int fId)
{
  if ((fId < 2) || (fId >= MaxOpenFiles+2)) {
    return((OpenFile *)0);
  }
  return((OpenFile*)(openFiles->Fetch(fId-2)));
}

//---------------------------------------------------------------------
// AddrSpace::UserStringToKernel
//    copy a string from an address space into
//    a kernel character buffer, which must be provided.
//
//    copying stops when a string terminator (\0) is found
//
//    At most MaxStringArgLength bytes will be copied
//
//   "vaddr" is the virtual address of the start of the string
//
//   "buf" points to a character buffer which will hold the string.
//
//   returns the length of the copied string if a string
//   terminator is found, -1 otherwise
//
//----------------------------------------------------------------------
int
AddrSpace::UserStringToKernel(int vaddr, char* buf)
{
  int length = 0;
  int dat;
  TranslationStatus status;

  DEBUG(dbgAddr, "Reading user string to kernel, starting virtual address: " << vaddr); 
  while(length < MaxStringArgLength) {
    status = ReadMem(vaddr,1,&dat);
    // n.b. this will fail if user string's page(s) are not in
    // memory.
    if (status != OK) {
      return(-1);
    }
    *buf = (char)dat;
    vaddr++;
    length++;
    if (*buf == '\0') break;
    buf++;
  }
  if (length >= MaxStringArgLength) {
    return(-1);
  } else {
    return(length);
  }
}

//---------------------------------------------------------------------
// AddrSpace::UserBufToKernel
//    copy a buffer of known length from an address space into
//    a kernel character buffer, which must be provided.
//
//   "vaddr" is the virtual address of the start of the buffer
//   "length" is its length
//
//   "buf" points to the target buffer
//
//   returns "length" on success, else -1
//----------------------------------------------------------------------
int
AddrSpace::UserBufToKernel(int vaddr, int length, char* buf)
{
  int knt;
  TranslationStatus status;
  int dat;

  DEBUG(dbgAddr, "Reading user buffer to kernel, starting virtual address: " << vaddr << " ,length " << length << " bytes."); 
  for(knt=0;knt<length;knt++) {
    status = ReadMem(vaddr,1,&dat);
    // n.b. this will fail if user buf's page(s) are not in
    // memory.
    if (status != OK) {
      return(-1);
    }
    *buf = (char)dat;
    vaddr++;
    buf++;
  }
  return(length);
}

//---------------------------------------------------------------------
// AddrSpace::KernelToUserBuf
//    copy a string of known length from a kernel into
//    an address space
//
//   "vaddr" is the virtual address of the start of the target buffer
//
//   "length" is the length of the buffer.
//
//   "buf" points to a kernel source buffer
//
//   returns "length" on success, else -1
//----------------------------------------------------------------------
int
AddrSpace::KernelToUserBuf(int vaddr, int length, char* buf)
{
  TranslationStatus status;
  int knt;

  DEBUG(dbgAddr, "Writing user buffer from kernel, starting virtual address: " << vaddr << " ,length " << length << " bytes."); 
  for(knt=0;knt < length;knt++) {
    status = WriteMem(vaddr,1,(int)(*buf));
    // n.b. this will fail if user buf's page(s) are not in
    // memory.
    if (status != OK) {
      return(-1);
    }
    vaddr++;
    buf++;
  }
  return(length);
}



