/* syscalls.h 
 * 
 * *** NOTE: Some Nachos error codes from errno.h are described here and are
 *     examples of error codes that could be used to making testing and
 *     debugging simpler because they can provide more information than
 *     just that a system call failed.
 *     You are not required to use or implement them.
 *     Preexisting system calls do NOT return specific error codes from errno.h.
 *
 * Wed Mar 10 10:54:00 EST 2004
 *
 * 	Nachos system call interface.  These are Nachos kernel operations
 * 	that can be invoked from user programs, by trapping to the kernel
 *	via the "syscall" instruction.
 *
 *	This file is included by user programs and by the Nachos kernel. 
 *
 * Copyright (c) 1992-1993 The Regents of the University of California.
 * All rights reserved.  See copyright.h for copyright notice and limitation 
 * of liability and disclaimer of warranty provisions.
 */

/*
 *  IMPORTANT!!!
 *
 *  This system call interface is used by CS350 assignment testing 
 *  scripts.  You are free to add new system calls to this interface.
 *  However:
 *     - DO NOT remove any existing system call
 *     - DO NOT change the signature (number and types of parameters)
 *           of any existing system call
 *     - DO NOT change the system call code of any existing system call
 *     - DO NOT change any of the existing #define definitions
 *
 *  Ignore this at your own peril.  If the marking scripts fail because
 *  you've changed the existing system call interface, then you will lose
 *  marks.
 */

#ifndef SYSCALLS_H
#define SYSCALLS_H

#include "copyright.h"
#include "errno.h"
/* system call codes -- used by the stubs to tell the kernel which system call
 * is being asked for
 * Note that in Nachos these don't need to be number consecutively.
 */

/* Process and thread related calls */
#define SC_Halt		        0
#define SC_Exit		        1
#define SC_Exec		        2
#define SC_ExecV	        3
#define SC_Join		        4
#define SC_SetPriority          5
#define SC_GetPriority          6
#define SC_GetId                7
#define SC_GetParentId          8
#define SC_ThreadFork	        9
#define SC_ThreadYield	       10
#define SC_ThreadExit          11
#define SC_Suspend             12
#define SC_Resume              13

/* File and File System related calls */
#define SC_Create	       20
#define SC_Remove              21
#define SC_Open		       22
#define SC_Read		       23
#define SC_Write	       24
#define SC_Close	       25
#define SC_Seek                26
#define SC_CreateDir           27
#define SC_RemoveDir           28
#define SC_GetFileLen          29

/* Lock related calls */
#define SC_LockOpen            40
#define SC_LockClose           41
#define SC_LockAcquire         42
#define SC_LockRelease         43

/* Shared Memory related calls */
#define SC_SharedMemOpen       50
#define SC_SharedMemClose      51

/* Pipe related calls */
#define SC_CreatePipe          60
#define SC_RemovePipe          61


/* Limit how long the name of a lock, file, or pathname can be */
#define MAX_NAME_AND_PATH_LEN  (128)

#ifndef IN_ASM

/* The system call interface.  These are the operations the Nachos
 * kernel needs to support, to be able to run user programs.
 *
 * Each of these is invoked by a user program by simply calling the 
 * procedure; an assembly language stub stuffs the system call code
 * into a register, and traps to the kernel.  The kernel procedures
 * are then invoked in the Nachos kernel, after appropriate error checking, 
 * from the system call entry point in exception.cc.
 */

/************** Processes and theads *************************************/

/* Stop Nachos, and print out performance stats */
void Halt();		
 

/* Process control operations: Exit, Exec, Execv, and Join */

/* A unique identifier for a Nachos process */
/* Nachos process identifiers are non-negative. */
typedef int ProcessId;	

/* This user program is done, status is the exit status code of
   the process.  status must be non-negative 
   status = 0 is used to indicate success
   status > 0 can be used to indicate failures */
void Exit(int status);	

/* Run the specified executable, with no args */
/* If ExecV is implemented, this can be implemented as a call to ExecV.
 * Returns a unique non-negative process identifier for the new
 *  process, or a negative value if the process could not
 *  be created.
 */ 
ProcessId Exec(char* exec_name);

/* Run the executable, stored in the Nachos file "argv[0]", with
 * parameters stored in argv[1..argc-1] and return a
 * unique non-negative process identifier for the new process.
 * Returns a negative value in case of failure.
 *
 * A few example error codes might be:
 *    NACHOS_ENAMETOOLONG path to executable uses too many characters
 *    NACHOS_EREACHEDLIMIT too many arguments to handle properly
 *    NACHOS_EINVAL one or more arguments is not valid
 */
ProcessId ExecV(int argc, char* argv[]);
 
/* Only return once the user program "id" has finished.  
 * Return the exit status.
 * A negative return value indicates that Join has
 *  failed.  (This means that all processes should return
 *  non-negative status values.) 
 *
 * An example error code might be:
 *    NACHOS_EINVAL the id argument is not valid
 */
int Join(ProcessId id); 	
 
/* Return the process id of the calling process */
ProcessId GetId();

/* Return the process id of the parent of the calling process.
   Return the process' own id if it has no parent, i.e., if it
   is the root process. */
ProcessId GetParentId();

/* Suspend and Resume child processes.  "childId" identifies the
 *  child process that is to be suspended or resumed.
 *  The suspended process should stop runing until it is resumed.
 *  It is an error for a process to attempt to suspend or resume
 *  processes other than its children.
 * These calls should return 0 for success, or a negative error
 *  code to indicate an error.
 *
 * An example error code might be:
 *    NACHOS_EINVAL the childId argument is not valid
 */
int Suspend(ProcessId childId);
int Resume(ProcessId childId);

/* Set thread scheduling priority
 *   Set the scheduling priority of the calling thread.
 *
 *   There are three possible priorities: normal and low and very low.
 *   A thread at priority i should only
 *   run if there are no higher priority (i+1 or higher) threads to run.
 *
 * This call should return 0 for success, or a negative error
 *   code to indicate an error.
 *
 * An example error code might be:
 *    NACHOS_EINVAL the priority argument is not valid
 */
typedef int Priority; 

#define P_NORMAL    (3)
#define P_LOW       (2)
#define P_VERY_LOW  (1)

int SetPriority(Priority p);

/* Returns the current priority (positive value) 
 * of the calling thread.
 * Returns negative error code on failure
 *
 */
int GetPriority();

/* User-level thread operations.
 * These allow multiple threads to run within a user program. 
 */

/* Fork a thread to run a procedure ("func") in the *same* address space 
 * as the current thread.
 * Return 0 on success, negative error code on failure
 */
int ThreadFork(void (*func)(int),int arg);

/* Yield the CPU to another runnable thread, whether in this address space 
 * or not. 
 */
void ThreadYield();		

/* Terminate the calling thread.
 * If the calling thread is the last thread in its process, this
 * call should also terminate the process, as if by a call to
 * Exit(0).
 * If the calling thread is not the last thread in its process,
 * the process should not be terminated.
 */
void ThreadExit();

/************** Files and File system *************************************/

/* File system operations: Create, Remove, Open, Read, Write, Close
 * These functions are patterned after UNIX -- files represent
 * both files *and* hardware I/O devices.
 *
 * Note that the Nachos file system has a stub implementation, which
 * can be used to support these system calls if the regular Nachos
 * file system has not been implemented.
 */
 
/* A unique identifier for an open Nachos file. */
typedef int OpenFileId;	

/* when a process starts up, it has two open files, representing 
 * keyboard input (ConsoleInId) and display output (ConsoleOutId).
 * These are similar to Unix stdin and stdout.
 * Read can be used directly on ConsoleInId and Write can be used
 * directly on ConsoleOutId, without first opening the console devices
 */

#define ConsoleInId	0  
#define ConsoleOutId	1  
 
/* Create a Nachos file, with name "name"
 * Note: Create does not open the file. 
 * Return 0 on success, negative error code on failure
 */
int Create(char *name);

/* Remove a Nachos file, with name "name"
 * Return 0 on success, negative error code on failure
 *
 * An example of some error codes might be:
 *   NACHOS_ENOENT the file is not found 
 *   NACHOS_ENAMETOOLONG the name uses too many characters
 *   NACHOS_EFAULT the address for name is not part of the address space
 */
int Remove(char *name);

/* Open the Nachos file "name", and return an "OpenFileId" that can 
 * be used to read and write to the file.
 *
 * Mode specifies if the file should be opened for reading, writing,
 * or both.
 * Valid modes are:
 *   "r"   : read only 
 *   "r+"  : read and write (i.e., read/write)
 *   "w"   : write only 
 *   "w+"  : read and write (i.e., read/write)
 *   "rw"  : read and write (i.e., read/write)
 *   "a"   : write only appending to end of the file
 *   "a+"  : read and write file position is at the end of the file
 *
 * An example of some error codes might be:
 *   NACHOS_ENOENT the file does not already exist (is not found)
 *   NACHOS_ENAMETOOLONG the name uses too many characters
 *   NACHOS_EFAULT the address for name is not part of the address space
 *   NACHOS_EINVAL mode is not a valid mode
 */
OpenFileId Open(char *name, char *mode);

/* Close the file, we're done reading and writing to it.
 * Return 0 on success, negative error code on failure
 *
 * An example error code might be:
 *    NACHOS_EINVAL the id argument is not valid
 */
int Close(OpenFileId id);

/* Write "size" bytes from "buffer" to the open file.
 * Return the number of bytes actually written on success.
 * On failure, a negative error code is returned.
 */
int Write(char *buffer, int size, OpenFileId id);

/* Read "size" bytes from the open file into "buffer".  
 * Return the number of bytes actually read -- if the open file isn't
 * long enough, or if it is an I/O device, and there aren't enough 
 * characters to read, return whatever is available 
 * A return code of 0 should be used to indicate end-of-file.
 * Negative return codes should be used to indicate errors.
 */
int Read(char *buffer, int size, OpenFileId id);

/* Determine the length (in bytes) of a specified file.
 * Returns a non negative integer length
 * on success and appropriate negative error code on failure
 */
int GetFileLen(OpenFileId id);

/* Move the current read/write position for the file handle by offset
 * bytes from the origin, where the origin is F_START, F_CURRENT, F_END
 * Return the (old) current file position, or a negative error
 * code.
 */
#define F_START 0
#define F_CURRENT 1
#define F_END -1
int Seek(OpenFileId id, int offset, int origin);


/* Create a new directory
 * Returns 0 on success and an appropriate negative error
 * value on failure.
 *
 * An example of some error codes might be:
 *   NACHOS_ENAMETOOLONG the name uses too many characters
 *   NACHOS_EEXIST the directory already exists 
 */
int CreateDir(char *pathname);

/* Remove the specified directory
 * Returns 0 on success and an appropriate negative error
 * value on failure.
 *
 * An example of some error codes might be:
 *   NACHOS_ENAMETOOLONG the name uses too many characters
 *   NACHOS_ENOENT the directory is not found 
 */
int RemoveDir(char *pathname);

/********************* Locks *************************************/

/* unique identifier for a Nachos lock */
typedef int LockId;
 
/* Open a Nachos lock called "name", creating one if such
   a lock does not already exist.  Return a LockId that can
   be used to refer to the lock.
   Legitimate LockIds are non-negative.
   Returns a negative LockId to indicate an error.
*/
LockId LockOpen(char *name);
 
/* Close an open Nachos lock.
 * Once a process has closed a lock, it should no longer be
 * possible for that process to acquire that lock 
 * (unless it is first re-opened).
 * Return 0 on success, negative error code on failure.
 */
int LockClose(LockId id);
 
/* Acquire an opened lock in the specified mode, which may
 * be one of shared (S) or exclusive (X).
 * Returns 0 on success, a negative error code in case of failure.
 */
#define S_MODE 0
#define X_MODE 1
int LockAcquire(LockId id, int mode);
                                                                                
/* Release an acquired lock.
 * Return 0 on success, negative error code in case of failure.
 */
int LockRelease(LockId id);


/********************* Shared Memory *************************************/

/* Create and/or open a region of shared memory using
 * the specified name (used as a handle to refer to the region).
 * The parameter bytes specifies the size of the shared region of memory.
 *
 */
void *SharedMemOpen(char *name, unsigned int bytes);

/*
 * Close a previously opened region of shared memory
 */
int SharedMemClose(char *name);

/********************* Pipes *************************************/
/* Create a named pipe.
 *
 * On success returns 0.
 * On failure returns an appropriate negative error value.
 *
 * Note that this only creates a pipe and it must be opened
 * before it can be read and/or written.
 *
 * An example of some error codes might be:
 *   NACHOS_ENAMETOOLONG the name uses too many characters
 *   NACHOS_EEXIST the name already exists 
 */
int CreatePipe(char *name);

/* Remove a named pipe.
 *
 * On success returns 0.
 * On failure returns an appropriate negative error value.
 *
 * An example of some error codes might be:
 *   NACHOS_ENAMETOOLONG the name uses too many characters
 *   NACHOS_ENOENT the pipe is not found 
 */
int RemovePipe(char *name);

/* -------------------------------------------------------------------- */
#endif /* IN_ASM */

#endif /* SYSCALL_H */
